/*
*  PM - A Pre-Make Utility
*  Written Wednesday 6/7/1989 by Todd Whitmore
*
*  This utility will search through specified files and create
*  the output needed to generate a "makefile". By default, this
*  will be displayed to the standard output, but the user can
*  redirect the output using the DOS ">" or ">>" redirection
*  facility.
*
*  The default action is to search for all .C files, read them,
*  and generate the following:
*
*     a list of objects to be linked to form the final target
*     a list of .c_v and .c pairs for PVCS
*     a list of .obj's to be generated from the .c's, and all includes
*
*
*  Please not that nested includes can only go about 17 or 18 deep
*  with this utility because of the limited number of file handles
*  available. Anthing over this level would probably blow out the
*  stack anyhow, because we are using recursion (yes, professor)
*  as the technique to handle the nested includes. Most sane
*  programmers rarely nest includes beyon three levels anyway. I
*  have not extended the effort to test more than this, nor do I
*  ever plan to. If you get a stack overflow, relink the program
*  and kick up the stack a bit using the /ST:xxxx option
*
*  This program makes no false pretenses about portability. Compile
*  it under Microsoft C 5.1 or later.
*
*  
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <dos.h>
#include <ctype.h>
#include <time.h>

extern  int main(int argc,unsigned char * *argv);
static  void near pascal process(unsigned char *path);
static  void near pascal cleanse(unsigned char *string);
static  void near pascal lookup(unsigned char *ifile);
static  int  near pascal isok(unsigned char ch);
static  void near pascal helpme(void );

static char result[512], *argument = "*.C", temp[20], build[20];
static int to_term = 1;
static int pass;
static char first_time = 1;

#define OBJ_FILES 0
#define CV_FILES  1
#define C_FILES   2


main(int argc, char *argv[])
{
     struct find_t find;
     int status, i;
     long t;

     if (!isatty(fileno(stdout)))
     {
          to_term = 0;
          fprintf(stderr,"\nPM - Premake Utility\n");
          fprintf(stderr,"Public Domain Makefile Generator For C Programs\n");
          fprintf(stderr,"Written by Todd O. Whitmore\n");
     }

     if (strnicmp(argv[1],"/H",2) == 0)
          helpme();

     getcwd(result,sizeof(result));
     time(&t);
     printf("#----------------------------------------------------------\n");
     printf("#      Makefile generated on %s",ctime(&t));
     printf("#      in directory %s\n",result);
     printf("#----------------------------------------------------------\n\n");

     if (argc < 2)
     {
          argv[1] = argument;
          argc++;
     }

     for (pass = OBJ_FILES; pass <= C_FILES; pass++)
     {
          printf("\n\n");

          for (i=1; i < argc; i++)
          {
               if (!to_term)
                    fprintf(stderr,"%-70.70s\r",argv[i]);

               status = _dos_findfirst(argv[i],_A_NORMAL,&find);

               while (!status)
               {
                    _splitpath(find.name,NULL,NULL,temp,NULL);

                    switch (pass)
                    {
                    case OBJ_FILES:
                         if (first_time)
                         {
                             first_time = 0;
                             printf("#--------------------------------------------------------------\n");
                             printf("# objects that make up target\n");
                             printf("%-20.20s  %s.OBJ ","OBJECTS = ",temp);
                         }
                         else
                         {
                             printf("\\\n%-20.20s  %s.OBJ "," ",temp);
                         }

                         break;               
                    case C_FILES:
                         sprintf(build,"%s.OBJ",temp);
                         printf("\n#--------------------------------------------------------------\n");
                         printf("# dependencies for file %s\n",find.name);
                         printf("%-20.20s : %s",build,find.name);
                         process(find.name);
                         printf("\n\n");
                         break;
                    case CV_FILES:
                         sprintf(build,"%s.C",temp);
                         printf("\n#--------------------------------------------------------------\n");
                         printf("# dependencies for file  %s.C_V\n",temp);
                         printf("%-20.20s : %s_V\n\n",find.name,find.name);
                         break;
                    }

                    status = _dos_findnext(&find);
               }

          }

     }

     if (!to_term)
          fprintf(stderr,"%-70.70s\n","Completed . . .");

     exit(0);

}


static void near pascal process(char *path)
{
     char *buffer;
     FILE *fp;

     if ((fp = fopen(path,"r")) == NULL)
     {
          printf(" \\\n%22.22s %s <= can\'t find!"," ",path);
          return;
     }

     buffer = malloc(200);

     if (buffer == NULL)
     {
          printf("Out of memory\n");
          exit(1);
     }

     while (fgets(buffer,200,fp) != NULL)
     {
          if (strncmp(buffer,"#include",8) == 0)
          {
               memcpy(buffer,buffer+8,strlen(buffer));
               cleanse(buffer);
               lookup(buffer);
               printf(" \\\n%22.22s %s"," ",result);
               process(result);
          }
     }

     free(buffer);
     fclose(fp);

}


static void near pascal cleanse(char string[])
{
     int inptr, outptr;

     strupr(string);

/* find first valid token */
     for (inptr=0; string[inptr] && !isok(string[inptr]); inptr++)
          ;

     for (outptr=0; string[inptr]; inptr++)
          if (isok(string[inptr]))
               string[outptr++] = string[inptr];
          else
               break;

     string[outptr] = 0;
}

static void near pascal lookup(char *ifile)
{
     memset(result,0,sizeof(result));
     _searchenv(ifile,"INCLUDE",result);

     if (!result[0])
          strcpy(result,ifile);

     strupr(result);
}

static int near pascal isok(char ch)
{
     if (ch == ':')
     {
         fprintf(stderr,"\n\n*********************************************\n");
         fprintf(stderr,"** Error - absolute drive name in #include **\n");
         fprintf(stderr,"*********************************************\n");
         abort();
     }

     if (isupper(ch) || 
       ch == '.'     ||
       ch == '\\'    ||
       ch == '/'     ||
       ch == '_')
     {
          return(1);
     }
     else
     {
          return(0);
     }
}

static void near pascal helpme()
{
     static char *help[] = 
       {
          "PM is a utility designed to remove the drudgery of creating makefiles.",
          "PM will search through the existing .C files in the current directory",
          "in order to find all of the include files that are used. PM makes use",
          "of the DOS environment variable INCLUDE in order to help locate the",
          "include file specified in the .C program. Only .C files in the current",
          "directory can be searched at present, but you can specify ranges of",
          "program names, such as AD*.C on the command line. You can specify as",
          "many arguments on the command line as needed. The default argument is"
          "*.C.\n",
          NULL
     }
     ;

     register int i;


     for (i=0; help[i] != NULL; i++)
          fprintf(stderr,"%s\n",help[i]);


     exit(0);

}
