/*
*  PMA - A Pre-Make Utility
*  Written Wednesday 6/7/1989 by Todd Whitmore
*
*  This utility will search through specified files and create
*  the output needed to generate a "makefile". By default, this
*  will be displayed to the standard output, but the user can
*  redirect the output using the DOS ">" or ">>" redirection
*  facility.
*
*  This program makes no false pretenses about portability. Compile
*  it under Microsoft C 5.1 or later. 
*
*  *** Assembler Version ***
*/

/*============== Standard Include Files ===============*/
#include <stdio.h>         /* stream file io           */
#include <stddef.h>        /* common constants & types */
#include <stdlib.h>        /* common library functions */
#include <fcntl.h>         /* file control definitions */
#include <io.h>            /* low-level file io        */
#include <sys\types.h>     /* compiler types           */
#include <sys\stat.h>      /* S_I* definitions         */
#include <dos.h>           /* dos functions & REGS     */
#include <bios.h>          /* MSC Bios functions       */
#include <malloc.h>        /* memory allocation        */
#include <ctype.h>         /* classification macros    */
#include <string.h>        /* string functions         */
#include <time.h>          /* time functions           */


extern  int main(int argc,unsigned char * *argv);
static  void near pascal process(unsigned char *path);
static  void near pascal cleanse(unsigned char *string);
static  void near pascal lookup(unsigned char *ifile);
static  int  near pascal isok(unsigned char ch);
static  void near pascal helpme(void );

static char result[512], *argument = "*.asm", temp[20], build[20];
static int to_term = 1;
static int pass;
static char first_time = 1;

#define OBJ_FILES 0
#define ASM_FILES 1


main(int argc, char *argv[])
{
     struct find_t find;
     int status, i;
     long t;

     if (!isatty(fileno(stdout)))
     {
          to_term = 0;
          fprintf(stderr,"\nPMA - Premake Utility\n");
          fprintf(stderr,"Public Domain Makefile Generator For ASM Programs\n");
          fprintf(stderr,"Written by Todd O. Whitmore\n");
          fprintf(stderr,"Generating makefile . . .\n");
     }

     if (strnicmp(argv[1],"/H",2) == 0)
          helpme();

     getcwd(result,sizeof(result));
     time(&t);
     printf("#-------------------------------------------------------------\n");
     printf("#      Makefile generated on %s",ctime(&t));
     printf("#      in directory %s\n",result);
     printf("#-------------------------------------------------------------\n\n");

     if (argc < 2)
     {
          argv[1] = argument;
          argc++;
     }

     for (pass = OBJ_FILES; pass <= ASM_FILES; pass++)
     {
          printf("\n\n");

          for (i=1; i < argc; i++)
          {
               if (!to_term)
                    fprintf(stderr,"%-70.70s\r",argv[i]);

               status = _dos_findfirst(argv[i],(_A_NORMAL|_A_ARCH),&find);

               while (!status)
               {
                    _splitpath(find.name,NULL,NULL,temp,NULL);

                    switch (pass)
                    {
                    case OBJ_FILES:
                         if (first_time)
                         {
                             first_time = 0;
                             printf("#--------------------------------------------------------------\n");
                             printf("# objects that make up target\n");
                             printf("%-20.20s  %s.OBJ ","OBJECTS = ",temp);
                         }
                         else
                         {
                             printf("\\\n%-20.20s  %s.OBJ "," ",temp);
                         }

                         break;               
                    case ASM_FILES:
                         sprintf(build,"%s.OBJ",temp);
                         printf("\n#--------------------------------------------------------------\n");
                         printf("# dependencies for file %s\n",find.name);
                         printf("%-20.20s : %s",build,find.name);
                         process(find.name);
                         printf("\n\n");
                         break;
                    }

                    status = _dos_findnext(&find);
               }

          }

     }

     if (!to_term)
          fprintf(stderr,"%-70.70s\n","Completed . . .");

     exit(0);

}


static void near pascal process(char *path)
{
     char *where;
     char *buffer;
     FILE *fp;

     if ((fp = fopen(path,"r")) == NULL)
     {
          printf(" \\\n%22.22s %s <= can\'t find!"," ",path);
          return;
     }

     buffer = malloc(200);

     if (buffer == NULL)
     {
          printf("Out of memory\n");
          exit(1);
     }

     while (fgets(buffer,200,fp) != NULL)
     {
          strlwr(buffer);
          where = strstr(buffer,"include");

          if (where != NULL && strchr(where,'.') != NULL)
          {
               strcpy(buffer,where+8);
               cleanse(buffer);
               lookup(buffer);
               printf(" \\\n%22.22s %s"," ",result);
               process(result);
          }
     }

     free(buffer);
     fclose(fp);

}


static void near pascal cleanse(char string[])
{
     int inptr, outptr;

     strupr(string);

/* find first valid token */
     for (inptr=0; string[inptr] && !isok(string[inptr]); inptr++)
          ;

     for (outptr=0; string[inptr]; inptr++)
          if (isok(string[inptr]))
               string[outptr++] = string[inptr];
          else
               break;

     string[outptr] = 0;
}

static void near pascal lookup(char *ifile)
{

     if (access(ifile,0) == 0)
     {
         strcpy(result,ifile);
     }
     else
     {
        memset(result,0,sizeof(result));
        _searchenv(ifile,"INCLUDE",result);

        if (!result[0])
             strcpy(result,ifile);
     }

     strupr(result);
}

static int near pascal isok(char ch)
{
     if (ch == ':')
     {
         fprintf(stderr,"\n\n*********************************************\n");
         fprintf(stderr,"** Error - absolute drive name in \'include\' **\n");
         fprintf(stderr,"*********************************************\n");
         abort();
     }

     if (isalnum(ch) || 
       ch == '.'     ||
       ch == '\\'    ||
       ch == '/'     ||
       ch == '_')
     {
          return(1);
     }
     else
     {
          return(0);
     }
}

static void near pascal helpme()
{
     static char *help[] = 
       {
          "PMA is a utility designed to remove the drudgery of creating makefiles.",
          "PMA will search through the existing .ASM files in the current directory",
          "in order to find all of the include files that are used. PMA makes use",
          "of the DOS environment variable INCLUDE in order to help locate the",
          "include file specified in the .ASM program. Only .ASM files in the current",
          "directory can be searched at present, but you can specify ranges of",
          "program names, such as SW*.ASM on the command line. You can specify as",
          "many arguments on the command line as needed. The default argument is"
          "*.ASM\n",
          NULL
     }
     ;

     register int i;


     for (i=0; help[i] != NULL; i++)
          fprintf(stderr,"%s\n",help[i]);


     exit(0);

}
